/*
 * Copyright (C) 1998
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <string.h>	/* for strcmp() */

#include "monitor.h"	/* for monitor{Enter,Exit}() */
#include "compiler.h"


/*
 * Compile and invoke the method.
 */
bool_t compileAndInvokeMethod(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee) {
  int key = obj_monitor(fieldclass(&mb->fb));
#ifdef COMPILE_DEBUG
  printf("\n");
  printf("compileAndInvokeMethod() is called:\n  %s#%s %s\n  key: %d\n",
	cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature, key);
  fflush(stdout);
#endif

  monitorEnter(key);

  /* compile the called method */
  if (!(mb->CompiledCode) &&
	!(mb->fb.access & (ACC_ABSTRACT | ACC_NATIVE))) {
#ifdef COMPILE_DEBUG
    printf("  now compiling.\n");  fflush(stdout);
#endif
    if (compileMethod(mb)) {
      /* fail to compile.. */
#ifdef COMPILE_DEBUG
      printf("  compilation failed...\n");  fflush(stdout);
#endif
      mb->invoker = (bool_t (*)(JHandle*,struct methodblock *,int,ExecEnv*))
		access2invoker(mb->fb.access);
      monitorExit(key);
      return mb->invoker(o, mb, args_size, ee);
    }
  }

#ifdef COMPILE_DEBUG
  printf("compileAndInvokeMethod(): now call invoker.\n");
  fflush(stdout);
#endif
  monitorExit(key);
  return mb->invoker(o, mb, args_size, ee);
}


/*
 * Invoke the compiled method.
 */
bool_t invokeCompiledMethod(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee) {
  JavaFrame *old_frame, *frame;
  int access;
  stack_item *native_vars;
#ifdef RUNTIME_DEBUG
  int runtime_debug;
#endif

  /* allocate local var. space */
	/* caution: This code segment must be placed at code head. */
  __asm__("leal  -4(%%esp),%%edx\n\t"
	  "subl  %1,%%esp\n\t"
	  "movl  %%edx,%0"
	  : "=m" (native_vars) : "r" (mb->nlocals * 4)
	  : "edx");
	/* native_vars = %esp - 4 */
	/* %esp -= mb->nlocals * 4 */

#ifdef RUNTIME_DEBUG
#  if 1
  runtime_debug = 1;
#  else
  if ((!strcmp(cbName(mb->fb.clazz), "java/lang/FloatingDecimal"))
	&& (!strcmp(mb->fb.name, "toJavaFormatString")))
    runtime_debug = 1;
  else if ((!strcmp(cbName(mb->fb.clazz), "java/util/Hashtable"))
	&& (!strcmp(mb->fb.name, "put")))
    runtime_debug = 1;
  else if ((!strcmp(cbName(mb->fb.clazz), "E2.E2_Algorithm"))
	&& (!strcmp(mb->fb.name, "blockEncrypt")))
    runtime_debug = 1;
  else
    runtime_debug = 0;
#  endif
#endif	/* RUNTIME_DEBUG */

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("\n");
    printf("invokeCompiledMethod() is called:\n  %s#%s %s\n",
	cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature);
    if (exceptionOccurred(ee)) {
      printf("  An exception not handled remains!\n");
    }
    fflush(stdout);

    showStackFrames(ee);
  }	/* if (runtime_debug) */
#endif	/* RUNTIME_DEBUG */

  old_frame = ee->current_frame;
  access = mb->fb.access;

  /* creates a new frame */
  if (access & ACC_SYNCHRONIZED)
    invokeSynchronizedJavaMethod(o, mb, args_size, ee);
  else
    invokeJavaMethod(o, mb, args_size, ee);

  frame = ee->current_frame;
  frame->lastpc = mb->code;
	/* lastpc is not initialized in invoke*JavaMethod() */

  /* call */

  /* restack from JVM stack to native stack */
  {
    struct methodblock *prev_method;

    prev_method = old_frame->current_method;
    if (!prev_method)  goto restack;
    if (prev_method->CompiledCode) {
      /* called by compiled method */
      frame->vars = (stack_item *)old_frame->returnpc;
      goto restack_done;
    }

    /* called by normal Java or native method */
  restack:
    {
      CodeInfo *info;
      char *argsizes;
      stack_item *args;

      info = (CodeInfo *)mb->CompiledCodeInfo;
      if (!info)  prepareCompiledCodeInfo(ee, mb);
	/* make arg. size table for restack */

      argsizes = info->argsizes;
      args = frame->vars;	/* points to JVM stack */

      frame->vars = native_vars;

      while (*argsizes) {
	if (*argsizes == 1) {
	  native_vars[0] = args[0];
	  native_vars--;  args++;
	}
	else {
	  native_vars[-1] = args[0];
	  native_vars[0] = args[1];
	  native_vars -= 2;  args += 2;
	}
	argsizes++;
      }	/* while (*argsizes) */
    }	/* restack: */

  restack_done:
  }

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("call compiled method.\n");
    printf("  o:0x%08x, mb:0x%08x, args_size:%d, ee:0x%08x\n",
	(int)o, (int)mb, args_size, (int)ee);
    printf("  mb->CompiledCode: 0x%08x\n", (int)mb->CompiledCode);
    fflush(stdout);

    {
      stack_item *vars = ee->current_frame->vars;
      char *sig;
      int i;

      if (args_size > 0) {
	printf("  args:\n");
	i = 0;
	if (!(access & ACC_STATIC)) {
	  printf("    0x%08x @ 0x%08x  ", (int)vars[0].h, (int)vars);
	  showObjectBody("L;", vars[0].h);
	  printf("\n");
	  i++;
	}
	sig = mb->fb.signature;
	if (sig[0] == '(')  sig++;

	for (; i < args_size; i++) {
	  JHandle *arg = vars[-i].h;
	  if ((sig[0] == 'D') || (sig[0] == 'J'))  i++;
	  printf("    0x%08x @ 0x%08x  ", (int)arg, (int)(vars - i));
	  sig = showObjectBody(sig, arg);
	  printf("\n");
	}
      }
      fflush(stdout);
    }
  }
#endif

#ifdef RUNTIME_DEBUG
  ((void (*) (JHandle *, struct methodblock *, int, ExecEnv *, int))
	mb->CompiledCode)(o, mb, args_size, ee, runtime_debug);
#else
  ((void (*) (JHandle *, struct methodblock *, int, ExecEnv *))
	mb->CompiledCode)(o, mb, args_size, ee);
#endif

  if (frame->monitor)
    monitorExit(obj_monitor(frame->monitor));
  ee->current_frame = frame->prev;

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("compiled method finished: %s#%s %s\n",
	cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature);
    fflush(stdout);

    printf("  free memory: handle %lld obj %lld\n",
		FreeHandleMemory(), FreeObjectMemory());
    fflush(stdout);

    if (exceptionOccurred(ee)) {
      JHandle *exc = ee->exception.exc;
      printf("  exception occurred:\n");
      printf("    0x%08x(%s)\n",
		(int)exc, cbName(exc->methods->classdescriptor));
      showExcStackTrace(exc);
      fflush(stdout);
    }
    printf("\n");
    fflush(stdout);
  }	/* if (runtime_debug) */
#endif


#ifdef RUNTIME_DEBUG
  runtime_debug = 0;
#endif

  return !(exceptionOccurred(ee));
}
